#include <string.h>
#include <stdlib.h>
#include <ctype.h>
#include <netcdf.h>

#include "ceil.h"
#include "nctools.h"
#include "files2stdin.h"
#include "version.h"

#define ERROR(msg,status) fprintf(stderr, "Error: %s\n", msg); exit(status)
#define WARNING(msg) fprintf(stderr, "Warning: %s\n", msg)

#define NO_SIGNAL -999.00
#define MAX_STRING_LENGTH 512

int ncstatus;
#define handle_error(a) if ((ncstatus = (a)) != NC_NOERR) { \
    fprintf(stderr, "NetCDF error on line %d: %s\n", __LINE__, \
            nc_strerror(ncstatus)); \
    exit(-1); \
}

static
void
add_string_attribute(int ncid, int fieldid, char *attname, char *value)
{
  if (value) {
    /* The +1 ensures that \0 is included at the end of each attribute
       in the netcdf file - is this necessary? */
    handle_error(nc_put_att_text(ncid, fieldid, attname,
				 strlen(value), value));
  }
}

static
void
add_attributes(int ncid, int fieldid, char *units,
	       char *units_html, char *long_name, char *standard_name)
{
  if (units) {
    handle_error(nc_put_att_text(ncid, fieldid, "units",
				 strlen(units), units));
  }
  if (units_html) {
    handle_error(nc_put_att_text(ncid, fieldid, "units_html",
				 strlen(units_html), units_html));
  }
  if (long_name) {
    handle_error(nc_put_att_text(ncid, fieldid, "long_name",
				 strlen(long_name), long_name));
  }
  if (standard_name) {
    handle_error(nc_put_att_text(ncid, fieldid, "standard_name",
                                 strlen(standard_name), standard_name));
  }
}

static
void
add_scale_factor(int ncid, int fieldid, float betacal)
{
  float cal = betacal * 1.0e-7;

  handle_error(nc_put_att_float(ncid, fieldid, "scale_factor",
				NC_FLOAT, 1, &cal));

  nct_append_attribute(ncid, fieldid, COMMENT_NAME,
       "This variable should be multiplied by the value in the scale_factor attribute.", "\n");
}

static
void
add_missing_value(int ncid, int fieldid, float value)
{
  handle_error(nc_put_att_float(ncid, fieldid,
				"missing_value",
				NC_FLOAT, 1, &value));
#ifndef CHIL_USE_OLD_NAMES
  handle_error(nc_put_att_float(ncid, fieldid,
				"_FillValue",
				NC_FLOAT, 1, &value));
#endif
}

static
void
add_missing_value_short(int ncid, int fieldid, short value)
{
  handle_error(nc_put_att_short(ncid, fieldid,
				"missing_value",
				NC_SHORT, 1, &value));
#ifndef CHIL_USE_OLD_NAMES
  handle_error(nc_put_att_short(ncid, fieldid,
				"_FillValue",
				NC_SHORT, 1, &value));
#endif
}

static
void
Usage(char **argv)
{
  fprintf(stderr, "Usage: %s [options] file1.dat [file2.dat ...]\n"
	  "       cat file1.dat [file2.dat ...] | %s [options]\n"
	  "Options (n = integer, f = float, s = string):\n"
	  "         -date YYYYMMDD output only data written on this date\n"
	  "      -averays n        average n rays\n"
	  "     -avegates n        average n gates\n"
	  "       -prefix dir      write netCDF files in specified directory\n"
	  "  -noisefactor f        set noise rejection threshold (default 4)\n"
	  "      -betacal f        multiply beta by f x 1.0e-7 to get correct value\n"
	  /*	  "    -elevation f        set elevation angle in degrees\n" */
	  "      -noclean          don't clean noisy rays\n"
	  "        -noraw          don't provide raw and uncleaned beta_raw parameter\n"
	  "         -user s        name and email address of the user\n"
	  "     -location s f f f  location of the lidar (e.g. Chilbolton), followed by\n"
	  "                          its latitude (N), longitude (E) and altitude (m)\n"
	  "         -site s        use known lat, lon and alt for site s - currently\n"
	  "                           understood sites are Chilbolton and Cabauw\n"
	  "      -comment s        comment to be added to each file\n"
	  "      -verbose          talk to me\n", argv[0], argv[0]);
  exit(0);
}

int
main(int argc,char **argv)
{
  LCTime time, last_time = {0, 0, 0, 0, 0, 0};
  LCCloudBases cloudbases;
  LCBackscatterProfile profile_hr,profile;
  char *tmp_name;
  char aname[MAX_STRING_LENGTH];
  char tmpbuf[MAX_STRING_LENGTH], *prefix=".";
  char units[] = "sr-1 m-1";
  char units_html[] = "sr<sup>-1</sup> m<sup>-1</sup>";
#ifdef CHIL_USE_OLD_NAMES
  char suffix[] = "lc";
#else
  char suffix[] = "ct75k";
#endif
  char join[2] = "\0\0";
  char *user = NULL;
  char *location = NULL;
  char *comment = NULL;
  char *llocation = NULL;
  float tmp_float, heights[3];
  float decimalhour, lastdecimalhour=-1, stdn, noisefactor=4.0;
  float azimuth = 0, elevation = 90, betacal = 1.0;
  float latitude, longitude, altitude;
  pid_t pid = 0;
  size_t index=0, tmp_index;
  size_t start[2], count[2];
  int averays=1, avegates=1, rays_read, cloudgate;
  int rayid, gateid, wlid, altid, latid, lonid;
  int ncid, rangeid, timeid, betaid, betarawid, stdnid;
  int betanoisyid, basesid, base1id, base2id, base3id;
  int elevationid, azimuthid, elevid;
  int dimids[2], ngates=0;
  int skipped_rays = 0;
  short int beta_raw[LC_MAX_GATES];
  short int tmp_word;
  unsigned short int iarg=0,r_year=0,r_month=0,r_day=0;
  char bydate=0;
  char verbose=0, file_open=0, quiet=0;
  char cleanray=1; /* 1=beta and beta_raw; 0=beta only (uncleaned); 2=beta only (cleaned) */
  char fixed_elevation = 0, fixed_azimuth = 0;
  char betacal_override = 0;

  /* Process command-line arguments */

  while (++iarg < argc) {
    if (*argv[iarg] != '-') {
      /* Found a ceilometer filename. The next function is cunning - it makes
       * stdin read the contents of a list of files. */
      if (!(pid = files2stdin(argc-iarg, argv+iarg))) {
	ERROR("Failed to read files", 1);
      }
      break;
    }
    else if (strcmp(argv[iarg],"-date") == 0 && iarg+1<argc) {
      bydate=1;
      if (sscanf(argv[++iarg],"%04hu%02hu%02hu",
		 &r_year,&r_month,&r_day) < 3) {
	ERROR("Failed to read date: should be in the form YYYYMMDD",1);
      }
    }
    else if (strcmp(argv[iarg],"-avegates") == 0 && iarg+1<argc) {
      avegates = atoi(argv[++iarg]);
      if (avegates < 1) {
	WARNING("Cannot average negative or zero gates");
	avegates = 1;
      }
    }
    else if (strcmp(argv[iarg],"-averays") == 0 && iarg+1<argc) {
      averays = atoi(argv[++iarg]);
      if (averays < 1) {
	WARNING("Cannot average negative or zero rays");
	averays = 1;
      }
    }
    else if (strcmp(argv[iarg],"-noisefactor") == 0 && iarg+1<argc) {
      noisefactor=atof(argv[++iarg]);
    }
    else if (strcmp(argv[iarg],"-betacal") == 0 && iarg+1<argc) {
      betacal = atof(argv[++iarg]);
      if (betacal <= 0.0) {
	ERROR("Multiplicative calibration factor cannot be negative or zero", 1);
      }
      betacal_override = 1;
    }
    else if (strcmp(argv[iarg],"-noclean") == 0) {
      cleanray=0;
    }
    else if (strcmp(argv[iarg],"-noraw") == 0) {
      cleanray=2;
    }
    else if (strcmp(argv[iarg],"-prefix") == 0 && iarg+1<argc) {
      prefix=argv[++iarg];
    }
    else if (strcmp(argv[iarg],"-user") == 0) {
      if (argc > iarg+1) {
	if (user) {
	  WARNING("User specified more than once");
	}
	user = argv[++iarg];
      }
      else {
	ERROR("User not specified",1);
      }
    }
    else if (strcmp(argv[iarg],"-comment") == 0) {
      if (argc > iarg+1) {
	if (comment) {
	  WARNING("One of the comments will not appear");
	}
	comment = argv[++iarg];
      }
      else {
	ERROR("Comment not specified", 1);
      }
    }
    else if (strcmp(argv[iarg],"-location") == 0) {
      if (argc > iarg+4) {
	if (location) {
	  WARNING("Location specified more than once");
	}
	location = argv[++iarg];
	latitude = atof(argv[++iarg]);
	longitude = atof(argv[++iarg]);
	altitude = atof(argv[++iarg]);
      }
      else {
	ERROR("Not all location arguments were provided", 1);
      }
    }
    else if (strcmp(argv[iarg], "-site") == 0) {
      if (argc > iarg+1) {
	++iarg;
	if (strcasecmp(argv[iarg], "Chilbolton") == 0) {
	  location = "Chilbolton";
	  latitude = CHIL_LAT;
	  longitude = CHIL_LON;
	  altitude = CHIL_ALT+4.0; /* 4 metres above ground level */
	}
	else if (strcasecmp(argv[iarg], "Cabauw") == 0) {
	  location = "Cabauw";
	  latitude = CABAUW_LAT;
	  longitude = CABAUW_LON;
	  altitude = CABAUW_ALT;
	}
	else {
	  WARNING("Site unknown: I only know about Chilbolton and Cabauw");
	}
      }
    }
    /* Elevation is in the file!!! If it's wrong use this option */
    else if (strcmp(argv[iarg], "-elevation") == 0) {
      if (argc > iarg+1) {
	elevation = atof(argv[++iarg]);
	fixed_elevation = 1;
      }
      else {
	ERROR("Elevation not specified",1);
      }
    }
    else if (strcmp(argv[iarg], "-azimuth") == 0) {
      if (argc > iarg+1) {
	azimuth = atof(argv[++iarg]);
	fixed_azimuth = 1;
      }
      else {
	ERROR("Azimuth not specified",1);
      }
    }
    else if (strcmp(argv[iarg],"-verbose") == 0) {
      verbose=1;
      quiet=0;
    }
    else if (strcmp(argv[iarg],"-quiet") == 0) {
      quiet=1;
      verbose=0;
    }

    else {
      fprintf(stderr, "Error: `%s' not understood or not accompanied "
	      "by enough subsequent arguments\n", argv[iarg]);
      Usage(argv);
    }
  }

  if (!location) {
    WARNING("Location unknown: latitude, longitude and altitude will not be written");
    llocation = "";
  }
  else {
    char *ch = strdup(location);
    llocation = ch;
    join[0] = '-';
    while (*ch) {
      *ch = tolower(*ch);
      ++ch;
    }
  }

  while (1) {
    rays_read = (LCGetAveragedRay(stdin,&time,&cloudbases,&profile_hr,
			       averays));
    if (rays_read) {
      /* PROCESS DATA */
      LCAverageRay(&profile_hr,&profile,avegates);
      if (cleanray) {
	if (cleanray == 1) {
	  /* First save raw profile */
	  memcpy(beta_raw, profile.beta, sizeof(*profile.beta)*profile.ngates);
	}
	LCCustomCleanRay(&cloudgate,&stdn,&profile,8,noisefactor);
      }
      decimalhour = ((float) time.hour) + ((float) time.minute)/60.0 +
	((float) time.second)/3600.0;
    }

    if (file_open && (!rays_read
		      || last_time.year < time.year
		      || last_time.month < time.month
		      || last_time.day < time.day)) {
      /* CLOSE FILE */
      if (verbose) {
	fprintf(stderr, "\n");
      }
      if (skipped_rays) {
	if (quiet) {
	  fprintf(stderr, "%s: ", aname);
	}
	if (skipped_rays == 1) {
	  snprintf(tmpbuf, MAX_STRING_LENGTH,
	   "Warning: 1 ray of raw data was omitted due to a clock reset.");
	}
	else {
	  snprintf(tmpbuf, MAX_STRING_LENGTH,
	   "Warning: %d rays of raw data were omitted due to a clock reset.",
		   skipped_rays);
	}
	tmpbuf[MAX_STRING_LENGTH-1] = '\0';
	fprintf(stderr, "%s\n", tmpbuf);
	handle_error(nc_redef(ncid));
	nct_append_attribute(ncid, NC_GLOBAL, COMMENT_NAME,
			     tmpbuf, "\n");
      }
      handle_error(nc_close(ncid));
      index = file_open = 0;
      if (bydate) {
	bydate = 2;
      }
    }
    else if (rays_read && lastdecimalhour >= decimalhour
	     && last_time.year == time.year 
	     && last_time.month == time.month
	     && last_time.day == time.day) {
      if (verbose) {
	fprintf(stderr, "X");
      }
      skipped_rays += rays_read;
      continue;
    }

    if (!rays_read || bydate == 2) {
      /* EXIT */
      exit(0);
    }

    if (!file_open && !(bydate && (r_year != time.year 
				  || r_month != time.month
				  || r_day != time.day))) {
      char *version = CHIL_VERSION;

      /* OPEN NEW FILE */
      snprintf(aname, MAX_STRING_LENGTH, 
	       "%s/%04hu%02hu%02hu_%s%s%s.nc",
	       prefix, time.year, time.month, time.day,
	       llocation, join, suffix);
      aname[MAX_STRING_LENGTH-1] = '\0';
      /* Open netCDF file for writing */
      handle_error(nc_create(aname, NC_CLOBBER, &ncid));
      skipped_rays = 0;
      index = 0;
      if (!quiet)
	fprintf(stderr, "Writing %s\n", aname);

      /* Add some global attributes */
      add_string_attribute(ncid, NC_GLOBAL, "Conventions", "CF-1.0");
      if (location) {
	handle_error(nct_append_attribute(ncid, NC_GLOBAL, "location",
					  location, " "));
      }

      tmp_name = "CT75K";
      handle_error(nc_put_att_text(ncid, NC_GLOBAL, "system",
				   strlen(tmp_name), tmp_name));

      if (location) {
	nct_append_attribute(ncid, NC_GLOBAL, "title", location, " ");
	nct_append_attribute(ncid, NC_GLOBAL, "source", location, " ");
      }
      tmp_name = "Vaisala 905-nm CT75K lidar ceilometer";
      nct_append_attribute(ncid, NC_GLOBAL, "title", tmp_name, " ");
      nct_append_attribute(ncid, NC_GLOBAL, "source", tmp_name, " ");

      nct_append_attribute(ncid, NC_GLOBAL, "source",
			   "Wavelength: 905 nm\n"
			   "Half-angle beam divergence: 0.75 mrad\n"
			   "Half-angle field of view: 0.66 mrad", "\n");

      add_string_attribute(ncid, NC_GLOBAL, "references", "Information on the data"
		   " is available at http://www.met.rdg.ac.uk/radar/doc/ct75k.html");

      tmp_word = time.day;
      handle_error(nc_put_att_short(ncid, NC_GLOBAL, "day", NC_SHORT,
				    1, &(tmp_word)));

      tmp_word = time.month;
      handle_error(nc_put_att_short(ncid, NC_GLOBAL, "month", NC_SHORT,
				    1, &(tmp_word)));

      tmp_word = time.year;
      handle_error(nc_put_att_short(ncid, NC_GLOBAL, "year", NC_SHORT,
				    1, &(tmp_word)));

      if (comment) {
	handle_error(nct_append_attribute(ncid, NC_GLOBAL, COMMENT_NAME,
					  comment, "\n"));
      }


      /* Add a `history' attribute */
      handle_error(nct_add_history(ncid, "NetCDF generated from original data", user));

      /* Add a `command_line' attribute */
      handle_error(nct_add_command_line(ncid, argc, argv));

      handle_error(nc_put_att_text(ncid, NC_GLOBAL, "software_version",
				   strlen(version), version));
 
      /* Add time/ray dimension */
      handle_error(nc_def_dim(ncid, "time", NC_UNLIMITED, &rayid));

      /* Add range/gate dimension  */
      ngates = profile.ngates;
      handle_error(nc_def_dim(ncid, "range", ngates, &gateid));

      /* Add some scalar variables */
      handle_error(nc_def_var(ncid, "wavelength", NC_FLOAT, 0, dimids, 
			      &wlid));
      add_attributes(ncid, wlid, "nm", NULL, "Lidar wavelength", NULL);

      if (location) {
	handle_error(nc_def_var(ncid, "latitude", NC_FLOAT, 0, dimids, 
				&latid));
	add_attributes(ncid, latid, "degrees_north", NULL,
		       "Latitude of lidar", "latitude");

	handle_error(nc_def_var(ncid, "longitude", NC_FLOAT, 0, dimids, 
				&lonid));
	add_attributes(ncid, lonid, "degrees_east", NULL,
		       "Longitude of lidar", "longitude");

	handle_error(nc_def_var(ncid, "altitude", NC_FLOAT, 0, dimids, 
				&altid));
	add_attributes(ncid, altid, "m", NULL,
		       "Height of instrument above mean sea level "
		       "(Ordnance Survey Great Britain)", NULL);
      }

      if (fixed_elevation) {
	handle_error(nc_def_var(ncid, "elevation", NC_FLOAT, 0, dimids,
				&elevationid));
	add_attributes(ncid, elevationid, "degrees", NULL,
		       "Elevation above horizon (fixed)", NULL);
      }
      if (fixed_azimuth) {
	handle_error(nc_def_var(ncid, "azimuth", NC_FLOAT, 0, dimids,
				&azimuthid));
	add_attributes(ncid, azimuthid, "degrees", NULL,
		       "Azimuth clockwise from due north (fixed)", NULL);
      }
      

      /* Add time coordinate variable */ 
      dimids[0] = rayid;
      handle_error(nc_def_var(ncid, "time", NC_FLOAT, 1, dimids,
			      &timeid));
      snprintf(tmpbuf, MAX_STRING_LENGTH,
	       "hours since %04d-%02d-%02d 00:00:00 +00:00",
	       time.year, time.month, time.day);
      tmpbuf[MAX_STRING_LENGTH-1] = '\0';
      add_attributes(ncid, timeid, tmpbuf, NULL,
		     "Decimal hours UTC since midnight", NULL);
      add_string_attribute(ncid, timeid, "axis", "T");

      /* Add range coordinate variable */
      dimids[0] = gateid;
      handle_error(nc_def_var(ncid, "range", NC_FLOAT, 1, dimids, &rangeid));
      add_attributes(ncid, rangeid, "km", NULL, "Range from lidar", NULL);
      add_string_attribute(ncid, rangeid, "axis", "Z");

      if (!fixed_elevation) {
	/* Add elev variable */
	dimids[0] = rayid;
	handle_error(nc_def_var(ncid, "elev", NC_FLOAT, 1, dimids,
				&elevid));
	add_attributes(ncid, elevid, "degrees", NULL,
		       "Elevation above horizon", NULL);
	add_missing_value(ncid, elevid, LC_NO_ELEVATION);
	nct_append_attribute(ncid, elevid, COMMENT_NAME,
	     "The instrument measures elevation in one direction only, "
	     "with a precision of 1 degree.", "\n");
      }

      /* Add cloud bases and stdn parameter */
      dimids[0] = rayid;
      handle_error(nc_def_var(ncid, "bases", NC_BYTE, 1, dimids,
			      &basesid));
      tmp_name = "Number of cloud bases";
      handle_error(nc_put_att_text(ncid, basesid, "long_name",
				   strlen(tmp_name), tmp_name));
      nct_append_attribute(ncid, basesid, COMMENT_NAME,
	   "The Vaisala algorithm for retrieving cloud base heights is "
	   "proprietary and details of it are not available.", "\n");

      handle_error(nc_def_var(ncid, "base1", NC_FLOAT, 1, dimids,
			      &base1id));
      add_attributes(ncid, base1id, "km", NULL,
		     "Range of the first cloud base from the lidar", NULL);
      add_missing_value(ncid, base1id, NO_SIGNAL);

      handle_error(nc_def_var(ncid, "base2", NC_FLOAT, 1, dimids,
			      &base2id));
      add_attributes(ncid, base2id, "km", NULL,
		     "Range of the second cloud base from the lidar", NULL);
      add_missing_value(ncid, base2id, NO_SIGNAL);

      handle_error(nc_def_var(ncid, "base3", NC_FLOAT, 1, dimids,
			      &base3id));
      add_attributes(ncid, base3id, "km", NULL,
		     "Range of the third cloud base from the lidar", NULL);
      add_missing_value(ncid, base3id, NO_SIGNAL);

      if (cleanray) {
	handle_error(nc_def_var(ncid, "stdn", NC_FLOAT, 1, dimids,
				&stdnid));
	add_attributes(ncid, stdnid, units, units_html,
		       "Standard deviation of the backscatter coefficient "
		       "background noise", NULL);
	add_scale_factor(ncid, stdnid, betacal);
      }

      /* Add backscatter coefficient */
      dimids[0] = rayid;
      dimids[1] = gateid;

      handle_error(nc_def_var(ncid, "beta", NC_SHORT, 2,
			      dimids, &betaid));
      add_attributes(ncid, betaid, units, units_html,
		     "Attenuated backscatter coefficient", NULL);

      add_scale_factor(ncid, betaid, betacal);

      betanoisyid = betaid;
      if (cleanray) {
	add_missing_value_short(ncid, betaid, 0);
	nct_append_attribute(ncid, betaid, COMMENT_NAME,
	     "Speckle noise has been removed, but note that when speckle noise is strong\n"
	     "its removal reduces the effective sensitivity of the instrument.", "\n");
	if (cleanray == 1) {
	  handle_error(nc_def_var(ncid, "beta_raw", NC_SHORT, 2,
				  dimids, &betarawid));
	  add_attributes(ncid, betarawid, units, units_html,
			 "Raw attenuated backscatter coefficient", NULL);
	  add_scale_factor(ncid, betarawid, betacal);
	  betanoisyid = betarawid;
	}
      }

      if (cleanray != 2) {
	nct_append_attribute(ncid, betanoisyid, COMMENT_NAME,
			     "The data can be affected by speckle "
			     "noise in cloud-free regions.", "\n");
      }

      if (betacal_override) {
	nct_append_attribute(ncid, NC_GLOBAL, COMMENT_NAME,
	     "Calibration has been performed by utilising the fact "
	     "that the integral of the attenuated backscatter coefficient "
	     "through thick stratocumulus is equal to 1/(2k),\nwhere k is "
	     "the extinction-to-backscatter ratio and is close to 18.75 "
	     "sr for liquid cloud droplets.\nWe have also taken multiple "
	     "scattering into account - for the Vaisala instrument it "
	     "has an effect of around 20%.\nThe scale_factor "
	     "attribute indicates the resulting calibration value. The "
	     "raw system calibration can be obtained by using a scale_factor "
	     "of 1.0e-7.", "\n");
      }
      else {
	nct_append_attribute(ncid, NC_GLOBAL, COMMENT_NAME,
	     "The scale_factor attribute of the beta variable represents "
	     "the system calibration figure.\nThis can be somewhat in error "
	     "so subsequent calibration is necessary if the backscatter "
	     "values are to be used quantitatively.", "\n");
      }

      if (comment) {
	nct_append_attribute(ncid, NC_GLOBAL, COMMENT_NAME, comment, "\n");
      }

      /* End define mode */
      handle_error(nc_enddef(ncid));

      /* Set the scalar variables */
      tmp_index = 0;
      tmp_float = LC_WAVELENGTH;
      handle_error(nc_put_var1_float(ncid, wlid, &tmp_index, &tmp_float));

      if (location) {
	handle_error(nc_put_var1_float(ncid, altid, &tmp_index, &altitude));
	handle_error(nc_put_var1_float(ncid, latid, &tmp_index, &latitude));
	handle_error(nc_put_var1_float(ncid, lonid, &tmp_index, &longitude));
      }

      if (fixed_elevation) {
	handle_error(nc_put_var1_float(ncid, elevationid,
				       &tmp_index, &elevation));
      }
      if (fixed_azimuth) {
	handle_error(nc_put_var1_float(ncid, azimuthid,
				       &tmp_index, &azimuth));
      }

      /* Set the range */
      for (tmp_index = 0; tmp_index < ngates; ++tmp_index) {
	tmp_float = profile.range[tmp_index] * 0.001;
	handle_error(nc_put_var1_float(ncid, rangeid,
				       &tmp_index, &tmp_float));
      }
      file_open = 1;
    }
   
    if (file_open) {
      /* WRITE RAY */
      handle_error(nc_put_var1_float(ncid, timeid, &index,
				     &decimalhour));
      if (!fixed_elevation) {
	handle_error(nc_put_var1_float(ncid, elevid, &index,
				       &(profile.elevation)));
      }
      handle_error(nc_put_var1_int(ncid, basesid, &index,
				   &(cloudbases.number)));

      heights[2] = heights[1] = heights[0] = NO_SIGNAL;
      if (cloudbases.number > 0) {
	heights[0] = cloudbases.height1 * 0.001;
	if (cloudbases.number > 1) {
	  heights[1] = cloudbases.height2 * 0.001;
	  if (cloudbases.number > 2) {
	    heights[2] = cloudbases.height3 * 0.001;
	  }
	}
      }

      handle_error(nc_put_var1_float(ncid, base1id, &index,
				   heights));
      handle_error(nc_put_var1_float(ncid, base2id, &index,
				   heights+1));
      handle_error(nc_put_var1_float(ncid, base3id, &index,
				   heights+2));

      if (cleanray) {
	handle_error(nc_put_var1_float(ncid, stdnid, &index,
				       &stdn));
      }
      start[0] = index;
      start[1] = 0;
      count[0] = 1;
      count[1] = ngates;
      handle_error(nc_put_vara_short(ncid, betaid, start,
				     count, profile.beta));
      if (cleanray == 1) {
	handle_error(nc_put_vara_short(ncid, betarawid, start,
				       count, beta_raw));
      }

      ++index;
      if (verbose) {
	fprintf(stderr, ".");
      }
    }
    lastdecimalhour=decimalhour;
    last_time = time;
  }
  exit(0);
}
